
[1, 2, 3].reduce(0, +)
[1, 2, 3].reduce(1, *)
[1, 2, 3].reduce(0, -)
["he", "ll", "o"].reduce("", +)

[1, 2, 3].reduce(0, { $0 + $1 })

// 'reduce' requires:
// - binary operator
// - somewhat-complicated type requirement
// - initial value

// "monoid"
// - binary operator
//   - 2 closure inputs, collection type, and 1 output type are ALL the same type
//   - associative
// - initial value

// (1+2+3+4+5+6+...)
//  OR EVEN BETTER, if it's a monoid, it's parallelizable
// (1+2)+(3+4)+...

// Can we make a 'reduce' that works with monoids more strictly? Goal:
//
//   [1, 2, 3].reduce(sum)
//

let sumTuple = (0, { (x: Int, y: Int) in x + y })

// reduce(
//   [E],
//   R,
//   (R, E) -> R
// ) -> R

extension Array {
    func reduce(monoid: (Element, (Element, Element) -> Element)) -> Element {
        return self.reduce(monoid.0, monoid.1)
    }
}

struct Monoid<A> {
    let unit: A
    let accumulate: (A, A) -> A
}

extension Array {
    func reduce(_ m: Monoid<Element>) -> Element {
        return self.reduce(m.unit, m.accumulate)
    }
}


let sum = Monoid<Int>(unit: 0, accumulate: +)
let product = Monoid<Int>(unit: 1, accumulate: *)

[1, 2, 3].reduce(sum)
[1, 2, 3].reduce(product)

let concat = Monoid<String>(unit: "", accumulate: +)

["h", "e", "l", "l", "o"].reduce(concat)

["h", "e", "l", "l", "o"]
    .map { $0.count }
    .reduce(sum)


infix operator &&&
func &&&(l: Bool, r: Bool) -> Bool { return l && r }
infix operator |||
func |||(l: Bool, r: Bool) -> Bool { return l || r }

let allTrue = Monoid<Bool>(unit: true, accumulate: &&&)
let anyTrue = Monoid<Bool>(unit: false, accumulate: |||)

[false, false, true].reduce(allTrue)
[true, true, true].reduce(allTrue)
[false, false, true].reduce(anyTrue)
[false, false, false].reduce(anyTrue)

[].reduce(allTrue)
[].reduce(anyTrue)

let isInvalid = ["Logan", "Moseley", "", "New York"]
    .map { $0.isEmpty }
    .reduce(anyTrue)

